
#include "GrowArray.h"
#include <stdlib.h>

int GA_Init( GrowArray * const ga, void (*pElementDestructor)(void *) ) {
  if( ga == NULL ) 
    return GA_NULL;
  ga->elements = NULL;
  ga->_hsize = 0;
  ga->_lsize = 0;
  ga->_hbound = -1;
  ga->_lbound = 0;
  ga->policy = MULTIPLICATIVE;
  ga->pElementDestructor = pElementDestructor;
  return GA_SUCCESS;
}


int GA_InitWithSize( GrowArray * const ga, const int hsize, const int lsize, void (*pElementDestructor)(void *) ) {
  if( ga == NULL )
    return GA_NULL;
  ga->_hsize = hsize;
  ga->_lsize = lsize;
  ga->_hbound = -1;
  ga->_lbound = 0;
  ga->elements = NULL;
  ga->policy = MULTIPLICATIVE;
  ga->pElementDestructor = pElementDestructor;
  if ((ga->_hsize + ga->_lsize) > 0) {
      ga->elements = (void ** )calloc( ga->_hsize + ga->_lsize, sizeof( void * ) );
      if (ga->elements == NULL) {
          perror( "GA_Init: out of memory" );
          ga->_hsize = ga->_lsize = 0;
          ga->_hbound = -1;
          ga->_lbound = 0;
	  return GA_OUT_OF_MEMORY;
      }
  }
  return GA_SUCCESS;
}

#if 0
int GA_Copy( GrowArray * const target, const GrowArray * const source ) {
  int i = 0;

  if( target == NULL || source == NULL )
    return GA_NULL;
  GA_Kill( target );
  target->_hsize = source->_hsize;
  target->_lsize = source->_lsize;
  target->_hbound = source->_hbound;
  target->_lbound = source->_lbound;
  target->policy = MULTIPLICATIVE;
  target->elements = NULL;
  if (( target->_hsize + target->_lsize) > 0) {
      target->elements = (void ** )malloc( sizeof( void * ) * (target->_hsize + target->_lsize) );
      if (target->elements == NULL) {
          perror( "GA_Copy: out of memory" );
          target->_hsize = target->_lsize = 0;
          target->_hbound = -1;
          target->_lbound = 0;
          return GA_OUT_OF_MEMORY;
      }
  }
  for (i = 0; i < (source->_hsize + source->_lsize); i++)
/*TODO: go from 0 to (-_lbound + _hbound) instead of _hsize + _lsize */
    target->elements[i] = source->elements[i];
  return GA_SUCCESS;
}
#endif

void GA_ClearAndKill( GrowArray * const ga ) {
  int i = 0;
  if( ga == NULL )
    return;
  if( ga->pElementDestructor != NULL ) {
    for (i = 0; i < (ga->_hsize + ga->_lsize); i++) {
      ( *(ga->pElementDestructor))( ga->elements[i] );
      free( ga->elements[i] );
    }
  }
  GA_Kill( ga );
}

void GA_Kill( GrowArray * const ga ) {
  if( ga == NULL )
    return;
  free (ga->elements);
  ga->elements = NULL;
  ga->_hsize = 0;
  ga->_lsize = 0;
  ga->_hbound = -1;
  ga->_lbound = 0;
}

void * GA_GetElement( const GrowArray * const ga, const int i ) {
  if( ga == NULL ) {
    perror( "GA_GetElement: GrowArray parameter was NULL" );
    return NULL;
  }
#ifdef GROW_ARRAY_ASSERT_ON_BOUNDS_ERROR
  assert((i >= ga->_lbound) && (i <= ga->_hbound));
#endif

  if (i < ga->_lbound || i > ga->_hbound) {
      perror( "GA_GetElement: index out of bounds" );
      return NULL;
  }
  return ga->elements[ ga->_lsize + i ];
}

int GA_SetElement( GrowArray * const ga, const int i, void * const newElement ) {
  void * curElement = ga->elements[ ga->_lsize + i ];
#ifdef GROW_ARRAY_ASSERT_ON_BOUNDS_ERROR
  assert((i >= ga->_lbound) && (i <= ga->_hbound));
#endif

  if (i < ga->_lbound || i > ga->_hbound) {
      perror( "GA_SetElement: index out of bounds" );
      return GA_ERROR;
  }
  if( ga->pElementDestructor != NULL )
    (*(ga->pElementDestructor))( curElement );
  free( curElement );
  curElement = newElement;
  return GA_SUCCESS;
}


int GA_AddH( GrowArray * const ga, void * const newElement ) {
  void ** newElements = NULL;

  if( ga == NULL )
    return GA_NULL;
  /* expand storage if necessary */
  if ( ga->_hbound == ( ga->_hsize - 1) ) {
    /* need to expand; do so according to expansion policy. */
    int newHSize;
    if ( ga->_hsize == 0)
      newHSize = 1;
    else {
      if ( ga->policy == MULTIPLICATIVE)
        newHSize = ga->_hsize * 2;
      else
        newHSize = ga->_hsize + 1;
    }

    /* allocate new storage and copy old elements into it */
    newElements = (void **) realloc( ga->elements, sizeof( void * ) * (newHSize + ga->_lsize ) );
    if (newElements == NULL) {
      perror( "GA_AddH: out of memory" );
      return GA_OUT_OF_MEMORY;
    }
    ga->_hsize = newHSize;
    ga->elements = newElements;
  }

  /* we now have enough room to add an element at the top of the array. */
  ga->elements[ga->_lsize + ++(ga->_hbound)] = newElement;
  return GA_SUCCESS;
}

#if 0
int GA_RemH( GrowArray * const ga ) {
  if( ga == NULL ) 
    return GA_NULL;
  if ( ga->_hbound < ga->_lbound) {
    perror( "GA_Remh: no elements to remove" );
    return GA_ERROR;
  }
  ga->_hbound--;
  return GA_SUCCESS;
}
#endif

#if 0
int GA_AddL( GrowArray * const ga, const void * const newElement) {
  void ** newElements = NULL;
  int i = 0;

  if( ga == NULL )
    return GA_NULL;
  /* expand storage if necessary */
  if ( ga->_lbound == -(ga->_lsize)) {
    /* need to expand; do so according to expansion policy. */
    int newLSize;
    if (ga->_lsize == 0)
      newLSize = 1;
    else {
      if ( ga->policy == MULTIPLICATIVE)
        newLSize = ga->_lsize * 2;
      else
        newLSize = ga->_lsize + 1;
    }

    /* allocate new storage and copy old elements into it */
    newElements = (void **) malloc( sizeof( void * ) * (newLSize + ga->_hsize) );
    if (newElements == NULL) {
        perror( "GA_AddL: out of memory" );
        return GA_OUT_OF_MEMORY;
    }

    for (i = 0; i <= ga->_hbound; i++)
      newElements[newLSize + i] = (ga->elements)[ga->_lsize + i];
    for (i = -1; i >= ga->_lbound; i--)
      newElements[newLSize + i] = (ga->elements)[ga->_lsize + i];

    /* delete old storage, and reassign to new */
    free( ga->elements );
    ga->elements = newElements;
    ga->_lsize = newLSize;
  }

  /* we now have enough room to add an element at the top of the array. */
  (ga->elements)[ga->_lsize + --(ga->_lbound)] = newElement;
  return GA_SUCCESS;
}

#endif

#if 0
int GA_RemL( GrowArray * const ga ) {
  if( ga == NULL ) 
    return GA_NULL;
  if ( ga->_lbound > ga->_hbound) {
    perror( "GA_RemL: no elements to remove" ); 
    return GA_ERROR;
  }

  ga->_lbound++;
  return GA_SUCCESS;
}
#endif

int GA_HBound( const GrowArray * const ga ) {
  if( ga == NULL ) 
    return GA_NULL;
  return ga->_hbound;
}

#if 0
int GA_LBound( const GrowArray * const ga ) {
  if( ga == NULL )
    return GA_NULL;
  return ga->_lbound;
}
#endif

#if 0 
int GA_NumElements( const GrowArray * const ga ) {
  if( ga == NULL)
    return GA_NULL ;
  return (ga->_hbound - ga->_lbound + 1);
}
#endif

#if 0
int GA_SetExpansionPolicy( GrowArray * const ga, const ExpansionPolicy newPolicy ) {
  if( ga == NULL ) 
    return GA_NULL;
  ga->policy = newPolicy;
  return GA_SUCCESS;
}
#endif
 
#if 0
ExpansionPolicy GA_GetExpansionPolicy( const GrowArray * const ga ) {
  if( ga == NULL )
    return GA_NULL;
  return ga->policy;
}
#endif
