/* Growable array template class. */

#ifndef __GROWARRAY_H
#define __GROWARRAY_H

#include <stdio.h>

/* Return values for the helper functions (see below): */
#define GA_SUCCESS 0
/* GA_ERROR is the generic error return code */
#define GA_ERROR -2
/* GA_NULL is returned if the a GrowArray parameters is NULL */
#define GA_NULL -3
#define GA_OUT_OF_MEMORY -4

/* Uncomment line below for assertions on bounds errors */
#define GROW_ARRAY_ASSERT_ON_BOUNDS_ERROR

#ifdef GROW_ARRAY_ASSERT_ON_BOUNDS_ERROR
#include <assert.h>
#endif

/* Dynamic array "class" whose internal storage can grow.
// Can add and remove elements at the high and low ends of the array.
// Keeps track of high and low bounds of the array.

// When initialized, defaults to 0 elements in array. Low bound
// defaults to 0; high bound defaults to -1. The first GA_Addh() will
// place an element at index 0; the first GA_Addl() will place an
// element at index -1.
*/

/* Memory expansion policy. Default is MULTIPLICATIVE. */
typedef enum {
  ADDITIVE,		/* Increase size by 1 when growing */
  MULTIPLICATIVE	/* Multiply size by 2 when growing */
} ExpansionPolicy;

/* The GrowArray "class" */
typedef struct {
  void ** elements;  /* array elements */
  /* Total size of the array is the sum of the next two sizes. */
  int _hsize;   /* how big the high part of the elements array is */
  int _lsize;   /* how big the low part of the elements array really is */
  int _hbound;  /* how many elements are filled on the high end */
  int _lbound;  /* how many elements are filled on the low end */
  ExpansionPolicy policy;  /* Memory expansion policy. See above. */
  void (*pElementDestructor)(void *);
} GrowArray;

/* GrowArray helper functions: 
   All return GA_SUCCESS on success and one of the other GA_ codes (see #defines above) on failure unless otherwise indicated. */ 

/* constructors */
int GA_Init( GrowArray * const newArray, void (*elementDestructor)(void *) );

/* if hsize + lsize > 0, creates space for and zeroes the right number of elements */
int GA_InitWithSize( GrowArray * const newArray, const int hsize, const int lsize, void (*elementDestructor)(void *) );

/* copy constructor/function:
-abandons all data in targetArray (i.e. doesn't free it)
-resized targetArray to be like sourceArray.
-copies each pointer in sourceArray->elements, in order, into targetArray->elements.
-targetArray->policy defaults to MULTIPLICATIVE regardless of sourceArray->policy */

#if 0
int GA_Copy( GrowArray * const targetArray, const GrowArray * const sourceArray );
#endif

/* GA_ClearAndKill is a thorough destructor:
-frees memory for everything stored in the array, using ElementDestructor.
-frees memory for freeThisArray->elements */
void GA_ClearAndKill( GrowArray * const freeThisArray );

/* GA_Kill is a minimal destructor:
-frees memory for freeThisArray->elements, but leaves any pointers stored
 in the array untouched
So only use GA_Kill if you have pointers external to freeThisArray pointing to
data also stored in the freeThisArray */
void GA_Kill( GrowArray * freeThisArray );

/* note: GA_{Get,Set}Element only access elements within bounds (do not expand the bounds of the array). */

/* returns the retrieved element, or NULL if it fails. */
void * GA_GetElement( const GrowArray * const array, const int i );

/* calls ElementDestructor on array[i], then sets array[i] = newElement */
int GA_SetElement( GrowArray * const array, const int i, void * const newElement );

/* adds an element to the high end of the array */
int GA_AddH( GrowArray * const array, void * const newElement );

/* removes the high element */
#if 0
int GA_RemH( GrowArray * const array );
#endif

/* adds an element to the low bound of the array */
#if 0 
int GA_AddL( GrowArray * const array, const void * const newElement );
#endif

/* removes the low element */
#if 0 
int GA_RemL( GrowArray * const array );
#endif

/* returns the high bound of the array */
int GA_HBound( const GrowArray * const array );

/* returns the low bound of the array */
#if 0
int GA_LBound( const GrowArray * const array );
#endif

/* returns the number of elements in the array */
#if 0
int GA_NumElements( const GrowArray * const array );
#endif

#if 0
int GA_SetExpansionPolicy( GrowArray * const array, const ExpansionPolicy newPolicy);
#endif

/* Returns the current ExpansionPolicy, or a negative number on failure */
#if 0
ExpansionPolicy GA_GetExpansionPolicy( const GrowArray * const array );
#endif

#endif  /* #defined __GROWARRAY_H_ */
