#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include "PNMImageOffsetable.h"

#define BUF_SIZE 1024
#define M_OFFSET "MOffset"
#define N_OFFSET "NOffset"

int alloc( PNMImageOffsetable * pPNM );
void clear( Image * data );

int PNM_Init( PNMImageOffsetable * pPNM ) {
  pPNM->mOffset = 0;
  pPNM->nOffset = 0;
  pPNM->xSize = 0;
  pPNM->ySize = 0;
  pPNM->depth = 3;
  pPNM->magic_number = -1;
  pPNM->plm_field_1 = 0;
  pPNM->plm_field_2 = 0.0;
  pPNM->data = NULL;
  return PNM_SUCCESS;
}

int PNM_InitFromFile( PNMImageOffsetable * pPNM, const char *filename) {
  FILE * fp = NULL;
  int numRead, row_index;
  unsigned int pixelr;
  int height, width, channels, magic_number;
  unsigned char newchar;


  if( pPNM == NULL )
    return PNM_NULL;
  PNM_Init( pPNM );

  fp = fopen(filename, "rb");
  if (fp == NULL)
    return PNM_FILE_OPEN_ERROR;

  if ( parseHeaderFrom( fp, &(pPNM->xSize), 
			&(pPNM->ySize),
			&(pPNM->depth),
			&(pPNM->mOffset),
			&(pPNM->nOffset),
			&(pPNM->magic_number),
			&(pPNM->plm_field_1),
			&(pPNM->plm_field_2) ) != PNM_SUCCESS )
    return PNM_FILE_ERROR;
fprintf(stderr, "READ HEADR OK\n");
  if( alloc( pPNM ) != PNM_SUCCESS )
      return PNM_ERROR;

  /* Read in the actual data */

  fread(&pixelr, 1, 1, fp);   /* cut off 10 that keeps showing up */

  height = pPNM->ySize;
  width = pPNM->xSize;
  channels = pPNM->depth;
  pPNM->certchans = 0;
  magic_number = pPNM->magic_number;

  if (magic_number == 2 || magic_number == 3 || magic_number == 5 || magic_number == 6) 
    {
      for(row_index = 0; row_index < height*width*channels; row_index++)
	{
	  fread(&newchar, sizeof(unsigned char), 1, fp);
	  pPNM->data[row_index] = (Image)newchar;
	}
      numRead = row_index;
    }
  else if (magic_number == 7 || magic_number == 8 || magic_number == 9 || magic_number == 10 ) {
    fprintf(stderr, "Attempt to %d*%d*%d=%d\n", width,height,channels, width*height*channels);
    numRead = fread(pPNM->data, sizeof(Image), width*height*channels, fp);
	fprintf(stderr, "Numread = %d\n",numRead);
  }
  else if( magic_number == 12 ){
    fprintf(stderr, "Attempt to %d*%d*(%d+1)=%d\n", width,height,channels, width*height*(channels+1));
    numRead = fread(pPNM->data, sizeof(Image), width*height*(channels+1)/2, fp);
    numRead+= fread(pPNM->data+numRead,sizeof(Image), width*height*(channels+1)/2, fp);
	fprintf(stderr, "Numread = %d\n",numRead);
  }
  else 
    {
      fprintf(stderr, "PNM_InitFromFile: could not load data, unknown format.\n");
      fclose(fp);
      return -1; 
    }
  
  fclose(fp);
  
  if(magic_number != 12 && (numRead != ( pPNM->xSize * pPNM->ySize * pPNM->depth)))
    return PNM_FILE_ERROR;
  else if( magic_number == 12 && numRead != (width*height*(channels+1)) )
    return PNM_FILE_ERROR;

  fprintf(stderr, "Read FILE OK\n\n");
  return PNM_SUCCESS;
}


int PNM_InitFromSize( PNMImageOffsetable * pPNM, int xSize, int ySize, int depth, int magic_number, int plm_field_1, float plm_field_2) {
  if (xSize <= 0 || ySize <= 0 || depth <= 0 || depth > 4 )
    return PNM_ERROR;
  PNM_Init( pPNM );

  pPNM->xSize = xSize;
  pPNM->ySize = ySize;
  pPNM->depth = depth;
  pPNM->magic_number = magic_number;
  pPNM->plm_field_1 = plm_field_1;
  pPNM->plm_field_2 = plm_field_2;

  if( alloc( pPNM ) != PNM_SUCCESS )
    return PNM_ERROR;
  return PNM_SUCCESS;
}

void PNM_Destroy(PNMImageOffsetable * pPNM) {
  clear( pPNM->data );
}

int PNM_Copy( PNMImageOffsetable * pTarget, const PNMImageOffsetable * pSource ){
  if( pTarget == NULL || pSource == NULL ) 
    return PNM_NULL;

  pTarget->mOffset = pSource->mOffset;
  pTarget->nOffset = pSource->nOffset;
  pTarget->xSize = pSource->xSize;
  pTarget->ySize = pSource->ySize;
  pTarget->depth = pSource->depth;
  pTarget->magic_number = pSource->magic_number;
  pTarget->plm_field_1 = pSource->plm_field_1;
  pTarget->plm_field_2 = pSource->plm_field_2;

  if (pSource->data != NULL) {
    if( alloc( pTarget ) != PNM_SUCCESS )
      return PNM_ERROR;
    memcpy( pTarget->data, pSource->data, pTarget->xSize * pTarget->ySize * pTarget->depth * sizeof(Image));
  }
  else
    pTarget->data = NULL;
  return PNM_SUCCESS;
}

int PNM_WriteToFile( const PNMImageOffsetable * pPNM, const char *filename) {
 
  int i;
  FILE *fp = fopen(filename, "wb");
  
  if (fp == NULL) {
    fprintf(stderr, "PNMImageOffsetable: Could not open file %s\n", filename);
    return PNM_FILE_OPEN_ERROR;
  }
  if(    pPNM->xSize <= 0 || 
         pPNM->ySize <= 0 || 
      !( pPNM->depth == 1 || pPNM->depth == 3) ||
         pPNM->data == NULL )
  {
    fprintf(stderr, "PNMImageOffsetable: Incomprehensible image paramters\n");
    fclose(fp);
    return PNM_ERROR;
  }

  if (pPNM->magic_number == 10)
    fprintf(fp, "PA\n");
  else
    fprintf(fp, "P%x\n", pPNM->magic_number);

  fprintf(fp, "# %s %d\n", M_OFFSET, pPNM->mOffset);
  fprintf(fp, "# %s %d\n", N_OFFSET, pPNM->nOffset);
  fprintf(fp, "%d %d\n255\n", pPNM->xSize, pPNM->ySize);


  fprintf(stderr, "channels: %d\n", pPNM->depth);

  /* Special things for plms */
  if ( pPNM->magic_number == 9 || pPNM->magic_number == 10 )
    {
      fprintf(fp, "%d\n%f\n", pPNM->plm_field_1, pPNM->plm_field_2);   
      if (fwrite( pPNM->data, sizeof(Image), pPNM->xSize * pPNM->ySize * pPNM->depth, fp) !=
	(size_t)( pPNM->xSize * pPNM->ySize * pPNM->depth)) 
      {
	fclose(fp);
	return PNM_FILE_ERROR;
      }

    }
  else if( pPNM->magic_number == 12 )
    {
      fprintf(fp, "%d\n%f\n", pPNM->plm_field_1, pPNM->plm_field_2);   
      if (fwrite( pPNM->data, sizeof(Image), pPNM->xSize * pPNM->ySize * (pPNM->depth+1), fp) !=
	(size_t)( pPNM->xSize * pPNM->ySize * (pPNM->depth+1))) 
      {
	fclose(fp);
	return PNM_FILE_ERROR;
      }

	}
  else if( pPNM->magic_number == 7 || pPNM->magic_number == 8 ) 
    {
      if (fwrite( pPNM->data, 
                  sizeof(Image), pPNM->xSize * pPNM->ySize * pPNM->depth, fp) !=
	     (size_t)( pPNM->xSize * pPNM->ySize * pPNM->depth)) 
      {
	fclose(fp);
	return PNM_FILE_ERROR;
      }
    }
  /*
  else if( pPNM->magic_number == 12 ) 
    {

				
    }
	*/
  else
    for (i=0; i < pPNM->xSize * pPNM->ySize * pPNM->depth; i++)
    {
      /* May have to change the value here to 255 for displaying if > 255 */ 
      if (pPNM->data[i] >= NaNChar) 
	fprintf(fp, "%c", (unsigned char)NaNChar);
      else
	fprintf(fp, "%c", (unsigned char)pPNM->data[i]);
    }

  fclose(fp);
  return PNM_SUCCESS;
}

int PNM_GetXSize(const PNMImageOffsetable * pPNM ) {
  if( pPNM == NULL ) 
    return PNM_NULL;
  return pPNM->xSize;
}

int PNM_GetYSize( const PNMImageOffsetable * pPNM ) {
  if( pPNM == NULL ) 
    return PNM_NULL;
  return pPNM->ySize;
}

int PNM_GetDepth( const PNMImageOffsetable * pPNM ) {
  if( pPNM == NULL ) 
    return PNM_NULL;
  return pPNM->depth;
}

int PNM_GetMOffset( const PNMImageOffsetable * pPNM ) {
  return pPNM->mOffset;
}

int PNM_GetNOffset( const PNMImageOffsetable * pPNM ) {
  return pPNM->nOffset;
}

Image * PNM_Pixel( const PNMImageOffsetable * pPNM, int x, int y) {
  if( pPNM == NULL )
    return NULL;
  assert( pPNM->data != NULL);
  assert(( x >= 0) && (x < pPNM->xSize));
  assert((y >= 0) && (y < pPNM->ySize));
  assert(&((pPNM->data)[(pPNM->depth) * (pPNM->xSize * y + x)])!=NULL);
  return &((pPNM->data)[(pPNM->depth) * (pPNM->xSize * y + x)]);
}

int PNM_SetMOffset(PNMImageOffsetable * pPNM, int mOffset) {
  if( pPNM == NULL )
    return PNM_NULL;
  pPNM->mOffset = mOffset;
  return PNM_SUCCESS;
}

int PNM_SetNOffset(PNMImageOffsetable * pPNM, int nOffset) {
  if( pPNM == NULL )
    return PNM_NULL;
  pPNM->nOffset = nOffset;
  return PNM_SUCCESS;
}

/* private functions: */

int alloc(PNMImageOffsetable * pPNM ) {
  clear( pPNM->data );
  assert( pPNM->xSize > 0);
  assert( pPNM->ySize > 0);
  assert( pPNM->depth > 0);
  if( pPNM->magic_number == 12 ) 
  pPNM->data = (Image * ) malloc( (sizeof( Image )) * pPNM->xSize * pPNM->ySize * (pPNM->depth+1) );
  else
  pPNM->data = (Image * ) malloc( (sizeof( Image )) * pPNM->xSize * pPNM->ySize * pPNM->depth );
  if( pPNM->data == NULL )
    return PNM_OUT_OF_MEMORY;
  return PNM_SUCCESS;
}

void clear( Image * data ) {
  if ( data ) {
    free( data );
    data = NULL;
  }
}

int parseHeaderFrom( FILE *fp, int * xSize, int * ySize, int *depth, int * mOffset, int * nOffset, int *magic_number, int * plm_field_1, float * plm_field_2 ) {
  int c;
  int junk;
  char * ptr;
  unsigned char buf[BUF_SIZE];

  /* - - GET INFO FROM PPM HEADER - - */
  fscanf(fp,"P%x", magic_number);

  /* - kill a comment - */
  memset(buf, 0, BUF_SIZE * sizeof(char));
  while(1)
    {
      c = getc(fp);
      if (c == '#')
	{  
	  if (fgets((char *) buf, BUF_SIZE, fp) == NULL) {
        fprintf(stderr, "PNMImageOffsetable: Error parsing comment\n");
	    return PNM_ERROR;
      }
	  ptr = (char *) buf;
	  /* Parse the comment; try to get M or N offset if possible */
	  while ((*ptr == '#') || (*ptr == ' '))
	    ptr ++;
	  if (!strncmp(ptr, M_OFFSET, strlen(M_OFFSET))) {
	    ptr += strlen(M_OFFSET);
	    if (sscanf(ptr, "%d", &(*mOffset)) != 1) {
          fprintf(stderr, "PNMImageOffsetable: Error parsing mOffset comment\n");
	      return PNM_FILE_ERROR;
        }
	  }
	  else if (!strncmp(ptr, N_OFFSET, strlen(N_OFFSET))) {
	    ptr += strlen(N_OFFSET);
	    if (sscanf(ptr, "%d", &(*nOffset)) != 1){
          fprintf(stderr, "PNMImageOffsetable: Error parsing nOffset comment\n");
	      return PNM_FILE_ERROR;
        }
	  }
       	}

      if (c==EOF) {
          fprintf(stderr, "PNMImageOffsetable: Unexpected EOF\n");
	      return PNM_FILE_ERROR;
	  }
      
      if (c>='0' && c<='9')
	{
	  /* rewind file pointer 1 char position */
	  fseek(fp, -1, SEEK_CUR); 
	  break;  
	}
    }
  
  fscanf(fp,"%d %d %d", xSize, ySize, &junk);
  
  /* Read off the two extra fields in plms and save for later. */
  if (*magic_number == 9 || *magic_number == 10 || *magic_number == 12)
    fscanf(fp,"%d %f", plm_field_1, plm_field_2);

  /* Calculate the depth. */
  if (*magic_number == 2 || *magic_number == 5 || *magic_number == 7 || *magic_number == 9)
    *depth = 1;
  else if (*magic_number == 3 || *magic_number == 6|| *magic_number == 8 || *magic_number == 10
				  || *magic_number == 12)
    *depth = 3;
	/* RGB + 1 Certainty */
  /*
  else if (*magic_number == 12 )
    *depth = 4; 
  */
  else {
    fprintf(stderr, "PNMImageOffsetable: Unknown pnm magic number\n");
    return PNM_ERROR;
  }
   
  /* Now we have all the parts parsed. */
  return PNM_SUCCESS;
}


