/* C port of the Matlab file est_qchirp2.m */    
/*
est_qchirp2.m                                          Steve Mann, 1992-1994
%
% estimate the 10 Taylor series (incl. qchirp) parameters from a pair of images
%
% The model is:   u = p1 + p2 x + p3 y + p4 x^2 +  p5 xy
%                 v = q6 + q7 x + q8 y + q9 y^2 + q10 xy
%
% Use: p = est_qchirp2(E1,E2);
%  or: [p1,p2,p3,p4,p5,p6,p7,p8,p9,p10] = est_qchirp2(E1,E2);
%  or : est_qchirp2(E1,E2);      % nicely formatted screen output
%
% sign convention: m (or X) down, n (or Y) to right
% properly handles situation when one or both images contain some NaN points
*/
#include<stdio.h>
#include<assert.h>
//uncomment these lines for testing time
//#include<unistd.h>
//#include<sys/time.h>
#include "mat_util.h"
#include "est_pseudo.h"
#include "corners2r.h"
#include "derivatives.h"

#include "par.h"

void pseudo_estimate(Var * parameters, Var * E1, Var * E2)
{
  Var * Dt, *Dm, *Dn;
  Var * DERIVATIVES;
  Var * Derivatives;
  Data p1, p2, p3, p4, p5, p6, p7, p8;

  /* - - alloc mem for structures Dt, Dm, Dn not returned - - */
  Dt = (Var *)calloc(1, sizeof(Var));
  Dm = (Var *)calloc(1, sizeof(Var));
  Dn = (Var *)calloc(1, sizeof(Var));

  DERIVATIVES = (Var *)calloc(1, sizeof(Var));
  Derivatives = (Var *)calloc(1, sizeof(Var));
  DERIVATIVES->channels = 1;
  Derivatives->channels = 1;
  parameters->data = (Data *)calloc(8, sizeof(Data));
  parameters->M = 8;   parameters->N = 1; 
  parameters->channels = E1->channels;
  DERIVATIVES->name = "DERIVATIVES";
  Derivatives->name = "Derivatives";

  derivatives(E1, E2, Dt, Dm, Dn);

#if COMPARISION_TESTING
  save_var_to_par(Dt, "Dt.par",  DOUBLE, "Hello World!");
  save_var_to_par(Dm, "Dm.par",  DOUBLE, "Hello World!");
  save_var_to_par(Dn, "Dn.par",  DOUBLE, "Hello World!");
#endif

  build_pseudo_sum_matrix(Dt, Dm, Dn, DERIVATIVES, Derivatives);
  div_mat_by_constant(DERIVATIVES, 4.0, DERIVATIVES);
  div_mat_by_constant(Derivatives, 2.0, Derivatives);

#if COMPARISION_TESTING
  save_var_to_par(DERIVATIVES, "DER.par",  DOUBLE, "DER");
  save_var_to_par(Derivatives, "Der.par",  DOUBLE, "Der");
#endif


#if DEBUG
  print(DERIVATIVES, 1, 1, 8, 5);
  print(Derivatives, 1, 1, 8, 1);
#endif
  
  gaussian_elimination(DERIVATIVES, parameters, Derivatives); 

  /* - - reorder, nornalize, and add the I matrix to the parameters - - */
  parameters->data[0] = parameters->data[0] * (Data)Dt->M;
  parameters->data[1] = parameters->data[1] + (Data)1.0;
  parameters->data[2] = parameters->data[2] * ((Data)Dt->N / (Data)Dt->M);
  parameters->data[3] = parameters->data[3] / (Data)Dt->M;

  parameters->data[4] = parameters->data[4] * (Data)Dt->N;
  parameters->data[5] = parameters->data[5] * (Data)Dt->M / (Data)Dt->N;
  parameters->data[6] = parameters->data[6] + (Data)1.0;
  parameters->data[7] = parameters->data[7] / (Data)Dt->N;

  p1 = parameters->data[3];
  p2 = parameters->data[1];
  p3 = parameters->data[2];
  p4 = parameters->data[7];
  p5 = parameters->data[5];
  p6 = parameters->data[6];
  p7 = parameters->data[0];
  p8 = parameters->data[4];

  parameters->data[0] = p1;
  parameters->data[1] = p2;
  parameters->data[2] = p3;
  parameters->data[3] = p4;
  parameters->data[4] = p5;
  parameters->data[5] = p6;
  parameters->data[6] = p7;
  parameters->data[7] = p8;

#if DEBUG
  printf("  pseudo parameters (BEFORE conversion):\n");
  print(parameters, 1, 1, 8, 1);
#endif

  parameters = pseudo2p(parameters);

  free(Dt->data); free(Dm->data); free(Dn->data);
  free(DERIVATIVES->data);  free(Derivatives->data);
  free(Dt); free(Dm); free(Dn);
  free(DERIVATIVES);   free(Derivatives);
}

/* ================================================================== */
void build_pseudo_sum_matrix(Var * Dt, Var * Dm, Var * Dn,
			    Var * DER,
			    Var * Der)
{
  int x, y, index, count, dummy;
  Var *m, *n, *mmDm_plus_mnDn, *mnDm_plus_nnDn;
  Var m_by_Dm, n_by_Dn, m_by_Dn, n_by_Dm;
  int chnls;
// the following are to measure time consumed (e.g. helped optimize code)
//  struct timeval tp;
//  int starttime,endtime;
//gettimeofday(&tp,NULL);
//starttime = tp.tv_usec + tp.tv_sec*1000000;


  chnls = Dt->channels;


  m_by_Dm.M = Dm->M; 
  m_by_Dm.N = Dm->N;
  m_by_Dm.channels = 1;
  m_by_Dm.var_type = INTERNAL_VARIABLE;
  m_by_Dm.data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data));
  /* small hack to get it to work with elemental multiply */
  m_by_Dm.data  = (Data *)m_by_Dm.data;  

  n_by_Dn.M = Dn->M; 
  n_by_Dn.N = Dn->N;
  n_by_Dn.channels = 1;
  n_by_Dn.var_type = INTERNAL_VARIABLE;
  n_by_Dn.data = (Data *)calloc(Dn->M * Dn->N * chnls, sizeof(Data));
  /* small hack to get it to work with elemental multiply */
  n_by_Dn.data  = (Data *)n_by_Dn.data;  
 
  m_by_Dn.M = Dn->M; 
  m_by_Dn.N = Dn->N;
  assert(Dm->M == Dn->M);
  assert(Dm->N == Dn->N);
  m_by_Dn.channels = 1;
  m_by_Dn.var_type = INTERNAL_VARIABLE;
  m_by_Dn.data = (Data *)calloc(Dn->M * Dn->N * chnls, sizeof(Data));
  /* small hack to get it to work with elemental multiply */
  m_by_Dn.data  = (Data *)m_by_Dn.data;  

  n_by_Dm.M = Dm->M; 
  n_by_Dm.N = Dm->N;
  n_by_Dm.channels = 1;
  n_by_Dm.var_type = INTERNAL_VARIABLE;
  n_by_Dm.data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data));
  /* small hack to get it to work with elemental multiply */
  n_by_Dm.data  = (Data *)n_by_Dm.data;  

  DER->data = (Data *)calloc(8 * 8 * chnls, sizeof(Data));
  DER->M = 8;  DER->N = 8;  DER->channels = 1; /* # channels */
  DER->var_type = INTERNAL_VARIABLE;
  
  n = (Var *)calloc(1, sizeof(Var));
  m = (Var *)calloc(1, sizeof(Var));
  mmDm_plus_mnDn = (Var *)calloc(1, sizeof(Var));
  mnDm_plus_nnDn = (Var *)calloc(1, sizeof(Var));

  n->data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data));
  m->data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data));
  mmDm_plus_mnDn->data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data));
  mnDm_plus_nnDn->data = (Data *)calloc(Dm->M * Dm->N * chnls, sizeof(Data)); 

  mmDm_plus_mnDn->name = "mmDm_plus_mnDn";
  mnDm_plus_nnDn->name = "mnDm_plus_nnDn";

  m->M = Dm->M;  m->N = Dm->N; 
  n->M = Dm->M;  n->N = Dm->N;
  mmDm_plus_mnDn->M = Dm->M;  mmDm_plus_mnDn->N =  Dn->N;
  mnDm_plus_nnDn->M =  Dm->M;  mnDm_plus_nnDn->N =  Dn->N;


  m->channels = 1;   m->var_type = INTERNAL_VARIABLE;  
  n->channels = 1;   n->var_type = INTERNAL_VARIABLE;
  mnDm_plus_nnDn->channels = chnls;
  mnDm_plus_nnDn->var_type = INTERNAL_VARIABLE;
  mmDm_plus_mnDn->channels = chnls;
  mmDm_plus_mnDn->var_type = INTERNAL_VARIABLE;


  /* - -  Initialize m, n, and some 'useful' variables - - */
  for(y=1, index = 0; y <= Dm->M; y++)
    {
      count = 0;
      dummy = 1;
      for(x=0; x < (Dm->N * chnls); x+= chnls, index++)
	{
	  m->data[index] = (Data)y;
	  n->data[index] = (Data)dummy;
	  if( (Dm->data[index] != NaNData) &&
	      (Dn->data[index] != NaNData) )
	    {
	      mmDm_plus_mnDn->data[index] = (Data)(m->data[index] *
						   m->data[index] *
						   Dm->data[index]) +
						     (Data)(m->data[index] *
							    n->data[index] *
							    Dn->data[index]);
	      mnDm_plus_nnDn->data[index] = (Data)(m->data[index] *
						   n->data[index] *
						   Dm->data[index]) +
						     (Data)(n->data[index] *
							    n->data[index] *
							    Dn->data[index]);
	      if(mmDm_plus_mnDn->data[index] == NaNData)
		mmDm_plus_mnDn->data[index] = NaNData - (Data)1;
	      if(mnDm_plus_nnDn->data[index] == NaNData)
		mnDm_plus_nnDn->data[index] = NaNData - (Data)1;
	    }
	  else
	    {
	      mmDm_plus_mnDn->data[index] = NaNData;
	      mnDm_plus_nnDn->data[index] = NaNData;
	    }
	  if(count == (chnls - 1))
	    {
	      count = 0;
	      dummy++;
	    }
	  else
	    count++;
	}
    }

#if COMPARISION_TESTING
  save_var_to_par(m, "m.par",  DOUBLE, "M matrix");
  save_var_to_par(n, "n.par",  DOUBLE, "N matrix");
  save_var_to_par(mmDm_plus_mnDn, "mm.par",  DOUBLE, "mmDm_plus_mnDn matrix");
  save_var_to_par(mnDm_plus_nnDn, "mn.par",  DOUBLE, "mnDm_plus_nnDn matrix");
#endif
  
  /* - - Remove NaN values so * here and in mat_util.c works - - */ 
  chval(Dt, (double)NaNData, 0.0); 
  chval(Dm, (double)NaNData, 0.0); 
  chval(Dn, (double)NaNData, 0.0);
  chval(mmDm_plus_mnDn, (double)NaNData, 0.0);  
  chval(mnDm_plus_nnDn, (double)NaNData, 0.0);  

  index = 0;   /* the following uses index to see where to put in the
                  matrix DER, and then increments index */


  /* pre calculate some commonly used matrices -James Dec 14,99*/
  elemental_multiply(m, Dm, &m_by_Dm); 
  elemental_multiply(n, Dn, &n_by_Dn);
  elemental_multiply(m, Dn, &m_by_Dn);
  elemental_multiply(n, Dm, &n_by_Dm);


if( chnls == GREY_SCALE ) {
  /* - - ROW 1 - - */
  multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, mmDm_plus_mnDn);

  //multiply_and_sum(&index, DER,  3, m, Dm, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, &m_by_Dm, mmDm_plus_mnDn);

  //multiply_and_sum(&index, DER,  3, n, Dm, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, &n_by_Dm, mmDm_plus_mnDn);

  multiply_and_sum(&index, DER,  2, Dm, mmDm_plus_mnDn);

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, mmDm_plus_mnDn);
  //multiply_and_sum(&index, DER,  3, m, Dn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, &m_by_Dn, mmDm_plus_mnDn);

  //multiply_and_sum(&index, DER,  3, n, Dn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, mmDm_plus_mnDn);

  multiply_and_sum(&index, DER,  2, Dn, mmDm_plus_mnDn);

  /* - - ROW 2 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, m, Dm);
  index+=1;

  // multiply_and_sum(&index, DER,  4, m, Dm, m, Dm);
  multiply_and_sum(&index, DER,  2, &m_by_Dm, &m_by_Dm);

  //multiply_and_sum(&index, DER,  4, n, Dm, m, Dm);
  //multiply_and_sum(&index, DER,  3, n, &m_by_Dm, Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dm, &m_by_Dm);

  //multiply_and_sum(&index, DER,  3, Dm, m, Dm);
  multiply_and_sum(&index, DER, 2, &m_by_Dm, Dm);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, m, Dm);
  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, &m_by_Dm);

  //multiply_and_sum(&index, DER,  4, m, Dn, m, Dm);
  //multiply_and_sum(&index, DER,  3, m, Dn, &m_by_Dm);
  multiply_and_sum(&index, DER,  2, &m_by_Dn, &m_by_Dm);

  //multiply_and_sum(&index, DER,  4, n, Dn, m, Dm);
  //multiply_and_sum(&index, DER,  3, n, Dn, &m_by_Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, &m_by_Dm);

  //multiply_and_sum(&index, DER,  3, Dn, m, Dm);
  multiply_and_sum(&index, DER,  2, Dn, &m_by_Dm);

  /* - - ROW 3 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, n, Dm);
  //multiply_and_sum(&index, DER,  4, m, Dm, n, Dm);
  index+=2;

  //multiply_and_sum(&index, DER,  4, n, Dm, n, Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dm, &n_by_Dm);

  //multiply_and_sum(&index, DER,  3, Dm, n, Dm);
  multiply_and_sum(&index, DER,  2, Dm, &n_by_Dm);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, n, Dm);
  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, &n_by_Dm);

  //multiply_and_sum(&index, DER,  4, m, Dn, n, Dm);
  //multiply_and_sum(&index, DER,  3, &m_by_Dm, Dn, n);
  multiply_and_sum(&index, DER,  2, &m_by_Dm, &n_by_Dn);

  //multiply_and_sum(&index, DER,  4, n, Dn, n, Dm);
  //multiply_and_sum(&index, DER,  3, &n_by_Dn, n, Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, &n_by_Dm);

  //multiply_and_sum(&index, DER,  3, n, Dn, Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, Dm);

  /* - - ROW 4 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, Dm);
  //multiply_and_sum(&index, DER,  3, m, Dm, Dm);
  //multiply_and_sum(&index, DER,  3, n, Dm, Dm);
  index+=3;
  multiply_and_sum(&index, DER,  2, Dm, Dm);

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, Dm);

  //multiply_and_sum(&index, DER,  3, m, Dn, Dm);
  multiply_and_sum(&index, DER,  2, &m_by_Dm, Dn);

  //multiply_and_sum(&index, DER,  3, n, Dn, Dm);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, Dm);

  multiply_and_sum(&index, DER,  2, Dn, Dm);

  /* - - ROW 5 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  3, m, Dm, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  3, n, Dm, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  2, Dm, mnDm_plus_nnDn);
  index+=4;

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  3, m, Dn, mnDm_plus_nnDn);
  multiply_and_sum(&index, DER,  2, &m_by_Dn, mnDm_plus_nnDn);

  //multiply_and_sum(&index, DER,  3, n, Dn, mnDm_plus_nnDn);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, mnDm_plus_nnDn);

  multiply_and_sum(&index, DER,  2, Dn, mnDm_plus_nnDn);

  /* - - ROW 6 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, m, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dm, m, Dn);
  //multiply_and_sum(&index, DER,  4, n, Dm, m, Dn);
  //multiply_and_sum(&index, DER,  3, Dm, m, Dn);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, m, Dn);
  index+=5;
  //multiply_and_sum(&index, DER,  4, m, Dn, m, Dn);
  //multiply_and_sum(&index, DER,  3, &m_by_Dn, m, Dn);
  multiply_and_sum(&index, DER,  2, &m_by_Dn, &m_by_Dn);

  //multiply_and_sum(&index, DER,  4, n, Dn, m, Dn);
  //multiply_and_sum(&index, DER,  3, &n_by_Dn, m, Dn);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, &m_by_Dn);

  //multiply_and_sum(&index, DER,  3, Dn, m, Dn);
  multiply_and_sum(&index, DER,  2, Dn, &m_by_Dn);

  /* - - ROW 7 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, n, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dm, n, Dn);
  //multiply_and_sum(&index, DER,  4, n, Dm, n, Dn);
  //multiply_and_sum(&index, DER,  3, Dm, n, Dn);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, n, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dn, n, Dn);
  index+=6;

  //multiply_and_sum(&index, DER,  4, n, Dn, n, Dn);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, &n_by_Dn);

  //multiply_and_sum(&index, DER,  3, Dn, n, Dn);
  multiply_and_sum(&index, DER,  2, &n_by_Dn, Dn);

  /* - - ROW 8 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, Dn);
  //multiply_and_sum(&index, DER,  3, m, Dm, Dn);
  //multiply_and_sum(&index, DER,  3, n, Dm, Dn);
  //multiply_and_sum(&index, DER,  2, Dm, Dn);

  //multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, Dn);
  //multiply_and_sum(&index, DER,  3, m, Dn, Dn);
  //multiply_and_sum(&index, DER,  3, n, Dn, Dn);
  index+=7;
  multiply_and_sum(&index, DER,  2, Dn, Dn);
}
else {
  /* - - ROW 1 - - */
  multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  3, m, Dm, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  3, n, Dm, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, Dm, mmDm_plus_mnDn);

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  3, m, Dn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  3, n, Dn, mmDm_plus_mnDn);
  multiply_and_sum(&index, DER,  2, Dn, mmDm_plus_mnDn);

  /* - - ROW 2 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, m, Dm);
  index+=3;
  multiply_and_sum(&index, DER,  4, m, Dm, m, Dm);
  multiply_and_sum(&index, DER,  4, n, Dm, m, Dm);
  multiply_and_sum(&index, DER,  3, Dm, m, Dm);

  multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, m, Dm);
  multiply_and_sum(&index, DER,  4, m, Dn, m, Dm);
  multiply_and_sum(&index, DER,  4, n, Dn, m, Dm);
  multiply_and_sum(&index, DER,  3, Dn, m, Dm);

  /* - - ROW 3 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, n, Dm);
  //multiply_and_sum(&index, DER,  4, m, Dm, n, Dm);
  index+=6;
  multiply_and_sum(&index, DER,  4, n, Dm, n, Dm);
  multiply_and_sum(&index, DER,  3, Dm, n, Dm);

  multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, n, Dm);
  multiply_and_sum(&index, DER,  4, m, Dn, n, Dm);
  multiply_and_sum(&index, DER,  4, n, Dn, n, Dm);
  multiply_and_sum(&index, DER,  3, Dn, n, Dm);

  /* - - ROW 4 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, Dm);
  //multiply_and_sum(&index, DER,  3, m, Dm, Dm);
  //multiply_and_sum(&index, DER,  3, n, Dm, Dm);
  index+=9;
  multiply_and_sum(&index, DER,  2, Dm, Dm);

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, Dm);
  multiply_and_sum(&index, DER,  3, m, Dn, Dm);
  multiply_and_sum(&index, DER,  3, n, Dn, Dm);
  multiply_and_sum(&index, DER,  2, Dn, Dm);

  /* - - ROW 5 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  3, m, Dm, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  3, n, Dm, mnDm_plus_nnDn);
  //multiply_and_sum(&index, DER,  2, Dm, mnDm_plus_nnDn);
  index+=12;

  multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, mnDm_plus_nnDn);
  multiply_and_sum(&index, DER,  3, m, Dn, mnDm_plus_nnDn);
  multiply_and_sum(&index, DER,  3, n, Dn, mnDm_plus_nnDn);
  multiply_and_sum(&index, DER,  2, Dn, mnDm_plus_nnDn);

  /* - - ROW 6 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, m, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dm, m, Dn);
  //multiply_and_sum(&index, DER,  4, n, Dm, m, Dn);
  //multiply_and_sum(&index, DER,  3, Dm, m, Dn);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, m, Dn);
  index+=15;
  multiply_and_sum(&index, DER,  4, m, Dn, m, Dn);
  multiply_and_sum(&index, DER,  4, n, Dn, m, Dn);
  multiply_and_sum(&index, DER,  3, Dn, m, Dn);

  /* - - ROW 7 - - */
  //multiply_and_sum(&index, DER,  3, mmDm_plus_mnDn, n, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dm, n, Dn);
  //multiply_and_sum(&index, DER,  4, n, Dm, n, Dn);
  //multiply_and_sum(&index, DER,  3, Dm, n, Dn);

  //multiply_and_sum(&index, DER,  3, mnDm_plus_nnDn, n, Dn);
  //multiply_and_sum(&index, DER,  4, m, Dn, n, Dn);
  index+=18;
  multiply_and_sum(&index, DER,  4, n, Dn, n, Dn);
  multiply_and_sum(&index, DER,  3, Dn, n, Dn);

  /* - - ROW 8 - - */
  //multiply_and_sum(&index, DER,  2, mmDm_plus_mnDn, Dn);
  //multiply_and_sum(&index, DER,  3, m, Dm, Dn);
  //multiply_and_sum(&index, DER,  3, n, Dm, Dn);
  //multiply_and_sum(&index, DER,  2, Dm, Dn);

  //multiply_and_sum(&index, DER,  2, mnDm_plus_nnDn, Dn);
  //multiply_and_sum(&index, DER,  3, m, Dn, Dn);
  //multiply_and_sum(&index, DER,  3, n, Dn, Dn);
  index+=21;
  multiply_and_sum(&index, DER,  2, Dn, Dn);
}

  for (x=0; x<DER->M; x++){
          for(y=x+1;y<DER->N;y++){
                DER->data[x+y*DER->M]=DER->data[x*DER->M+y];
          }
  }

   /* - - - - - BUILD vector Derivatives - - - - - */
  index = 0;
  Der->data = (Data *)calloc(8 * 1, sizeof(Data));
  Der->M = 8;  Der->N = 1;  Der->channels = 1; 
  Der->var_type = INTERNAL_VARIABLE;

  multiply_and_sum(&index, Der,  2, Dt, mmDm_plus_mnDn);
  multiply_and_sum(&index, Der,  3, Dt, m, Dm);
  multiply_and_sum(&index, Der,  3, Dt, n, Dm);
  multiply_and_sum(&index, Der,  2, Dt, Dm);

  multiply_and_sum(&index, Der,  2, Dt,  mnDm_plus_nnDn);
  multiply_and_sum(&index, Der,  3, Dt, m, Dn);
  multiply_and_sum(&index, Der,  3, Dt, n, Dn);
  multiply_and_sum(&index, Der,  2, Dt, Dn);

  /* - - need to negate Derivatives - - */
  Der->data[0] = -(Der->data[0]);
  Der->data[1] = -(Der->data[1]);
  Der->data[2] = -Der->data[2];
  Der->data[3] = -Der->data[3];
  Der->data[4] = -Der->data[4];
  Der->data[5] = -Der->data[5];
  Der->data[6] = -Der->data[6];
  Der->data[7] = -Der->data[7];

  /* - - free temporary Variable memory - - */
  free(m->data);    free(m); 
  free(n->data);    free(n); 
  free(mnDm_plus_nnDn->data);  free(mnDm_plus_nnDn);
  free(mmDm_plus_mnDn->data);  free(mmDm_plus_mnDn);
  free(m_by_Dm.data);   
//  gettimeofday(&tp,NULL);
//endtime = tp.tv_usec + tp.tv_sec*1000000;
//printf("Time for build_pseudo_matrix %d usecs\n",abs(endtime-starttime));

}


/* ================================================================== */
Var * pseudo2p(Var * p)
{
  int index;
  Data x[4] = {0.0, 0.0, 1.0, 1.0};
  Data y[4] = {0.0, 1.0, 0.0, 1.0};
  Data extra;
  Data det;
  Data u[4], v[4];
  Data corners[8];
  Data ptmp[9];

  for(index = 0; index < 4; index++)
    {
      u[index] = (p->data[0]) +
	((p->data[1]) * (x[index])) +
	  ((p->data[2]) * (y[index])) + 
	    ((p->data[6]) * (x[index]) * (x[index])) +
	      ((p->data[7]) * (x[index]) * (y[index]));
    }    

  for(index =0; index < 4; index++)
    v[index] = p->data[3] +
                 p->data[4] * x[index] +
		   p->data[5] * y[index] + 
		     p->data[6] * x[index] * y[index] +
		       p->data[7] * y[index] * y[index];

  corners[0] = u[0];
  corners[1] = v[0];
  corners[2] = u[1];
  corners[3] = v[1];
  corners[4] = u[2];
  corners[5] = v[2];
  corners[6] = u[3];
  corners[7] = v[3];

  corners2r(p->data, corners);

  /* - - take the inverse - - */
  det = (p->data[0] * ((p->data[4]) - (p->data[7] * p->data[5]))) -
	   (p->data[1] * ((p->data[3]) - (p->data[6] * p->data[5]))) +
	     (p->data[2] *
	      ((p->data[3] * p->data[7]) - (p->data[6] * p->data[4])));
  
  /* - - copy values into array so we don't overwrite -- */
  ptmp[0] = p->data[0];
  ptmp[1] = p->data[1];
  ptmp[2] = p->data[2];
  ptmp[3] = p->data[3];
  ptmp[4] = p->data[4];
  ptmp[5] = p->data[5];
  ptmp[6] = p->data[6];
  ptmp[7] = p->data[7];

  
  p->data[0] = ((ptmp[4] * 1.0) -
		 (ptmp[7] * ptmp[5])) / det;
  p->data[1] = ((ptmp[1] * 1.0) -
		 (ptmp[7] * ptmp[2])) / -(det);
  
  p->data[2] = ((ptmp[1] * ptmp[5]) -
		 (ptmp[4] * ptmp[2])) / det; 

  p->data[3] = ((ptmp[3] * 1.0) -
		 (ptmp[6] * ptmp[5])) / -det; 

  p->data[4] = ((ptmp[0] * 1.0) -
		 (ptmp[6] * ptmp[2])) / det; 
  p->data[5] = ((ptmp[0] * ptmp[5]) - 
		 (ptmp[3] * ptmp[2])) / -det; 
  p->data[6] = ((ptmp[3] * ptmp[7]) - 
		 (ptmp[4] * ptmp[6])) / det; 
  p->data[7] = ((ptmp[0] * ptmp[7]) -
		 (ptmp[6] * ptmp[1])) / -det; 
  extra = (((double)ptmp[0] * (double)ptmp[4]) -
	   ((double)ptmp[3] * (double)ptmp[1])) / (double)det;

  p->data[0] = (p->data[0]) / (extra);
  p->data[1] = p->data[1] / extra;
  p->data[2] = p->data[2] / extra;
  p->data[3] = p->data[3] / extra;
  p->data[4] = p->data[4] / extra;
  p->data[5] = p->data[5] / extra;
  p->data[6] = p->data[6] / extra;
  p->data[7] = p->data[7] / extra;

  return(p);
}

