#include <string.h>
#include "GrowArray.h"
#include <stdio.h>
#include "mat_util.h"
#include "pinverse.h"
#include "motion.h"
#include <errno.h>

int VERBOSE = 0;

void usage() {
  printf( "usage: pintegrate [input parameters filename] [reference frame "
          "number]\n"
          " Reads in a set of pairwise pchirp parameters (8 numbers per\n" 
          " entry) generated by estpchirp2m and modifies them so the\n" 
          " reference frame has the identity transformation. The\n" 
          " outputted list is printed to stdout and is one entry longer\n" 
          " than the input list.\n" );
}

Var * createDataVar(int m, int n) {
  Var *var = (Var *) malloc(sizeof(Var));
  init_var( var, NULL, m, n, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  memset(var->data, 0, m * n * sizeof(Data));
  return var;
}

Var * copyDataVar(const Var *orig) {
  Var *var = (Var *) malloc(sizeof(Var));
  var->name = NULL;
  var->data = NULL;
  var->var_type = INTERNAL_VARIABLE;
  var->channels = GREY_SCALE;
  var->M = orig->M;
  var->N = orig->N;
  var->data = (Data *) malloc(var->M * var->N * sizeof(Data));
  memcpy(var->data, orig->data, var->M * var->N * sizeof(Data));
  return var;
}

void readParamFile(GrowArray * pGA, const char *fn) {
/* Takes a GrowArray that has already been initialized (with GA_Init) and
   reads the contents of fn into it */

  float tmp;
  FILE *fp = fopen(fn, "r");
  int i = 0;
  Var * var;

  if (fp == NULL) {
    fprintf( stderr, "pintegrate: Error opening file %s: %s\n", fn, strerror( errno) );
    exit(1);
  }

  while ( fscanf(fp, "%f", &tmp) == 1 ) { /* there are lines left in the file */
    /* Create a new Var */
    var = createDataVar(8, 1);
    GA_AddH( pGA, var);
    var->data[0] = tmp;
    /* 7 more to go for this Var */
    for (i = 0; i < 7; i++) {
      if (fscanf(fp, "%f", &tmp) != 1) {
        fprintf( stderr, "pintegrate: Error reading file %s: %s\n", fn, strerror( errno ) );
        exit(1);
      }
      var->data[1+i] = tmp;
    } /* for */
  } /* while */
  fclose(fp);
} /* readParamFile */

int main(int argc, char **argv) {
  char *inFile = argv[1];
  int refFrame = atoi(argv[2]);
  GrowArray params, cum;
  int i = 0, j = 0;
  Var * identity = NULL;
  Var * cur = NULL;
  Var * pRef = NULL;
  Var * var = NULL;
  Var * pRefInv = NULL;
  Var * pTmp = NULL;

  if (argc != 3) {
    usage();
    return 1;
  }
  
  /* Read in the parameters */
  GA_Init( &params, Var_Destructor );
  readParamFile( &params, inFile );

  /* pretty sure these lines are of no use and bug prone for scripts
  printf("The in file for parameters: %s\n", inFile);
  printf("-----------------------------------------\n");
  */

  if( GA_HBound( &params ) < 0 ) {
    perror ( "pintegrate: No input parameters read" );
    return 1;
  }
  if ( (refFrame < 0) || (refFrame > GA_HBound( &params )+1 ) ) {
    fprintf( stderr, "pintegrate: Reference frame must be between 0 and %d\n", GA_HBound( &params )+1 );
    return 1;
  }

  /* Initialize with the identity and the first frame's parameters */
  GA_Init( &cum, Var_Destructor );
  identity = createDataVar(8, 1);
  identity->data[0] = identity->data[4] = 1;
  identity->data[1] = identity->data[2] = identity->data[3] = identity->data[5] = identity->data[6] = identity->data[7] = 0;
  GA_AddH( &cum, identity );
  identity = NULL;
  GA_AddH( &cum, copyDataVar( (Var * )GA_GetElement( &params, 0 ) ) );

  for (i = 1; i <= GA_HBound( &params ); i++) {
    cur = createDataVar(8, 1);
    pcompose( ( Var * ) GA_GetElement( &params, i ), ( Var * ) GA_GetElement( &cum, i ), cur);
    GA_AddH( &cum, cur);
  }

  /* to use another frame as reference, invert all values with 
     inverse of desired ref. frames' parameters */ 
  if (refFrame > 0) {
    pRef = copyDataVar( (Var *) GA_GetElement( &cum, refFrame ));
    pRefInv = createDataVar(8, 1);
    pinverse(pRef, pRefInv);
    for (i = 0; i <= GA_HBound( &cum ); i++) {
      pTmp = createDataVar(8, 1);
      pcompose( (Var *)GA_GetElement( &cum, i ), pRefInv, pTmp);
      /* A hack that gets around possible bug in SetElement? */
      for (j = 0; j < 8; j++)
        printf("%.6f ", pTmp->data[j]);
      printf("\n");
      GA_SetElement( &cum, i, pTmp );
    }
    // we'll live with a memory leak (until pintegrate quits) for now, to
    // quickfix this bug (needs more analysis later) --s
    //Var_Destructor( pRef );
    //Var_Destructor( pRefInv );
  }
  else {
    pRef = pRefInv = pTmp = NULL;
    /* Display result to stdout */
    for (i = 0; i <= GA_HBound( &cum ); i++) {
      var =(Var *) GA_GetElement( &cum, i );
      for (j = 0; j < 8; j++)
        printf("%.6f ", var->data[j]);
      printf("\n");
    }
  }
  /* A hack that gets around a seg-fault in the destructor. */
/*
  for (i=0; i<params._hsize * params._lsize; i++)
    free (params.elements[i]);
  GA_Kill( &params );
  for (i=0; i<cum._hsize * cum._lsize; i++)
    free (cum.elements[i]);
  GA_Kill( &cum );
*/
  return 0;
}
