
/*  
 *  This is a SIMPLE interface to the png library
 *  for writing RGBA images.  the four bytes are
 *  packed into an assumed 32-bit int in the RGBA
 *  order.
 *
 *  So we need the 
 *    filename, raw data ptr, width, height,
 *  and optional 
 *    comment (may be NULL)
 *    moffset, noffset (may be 0)
 *
 *  you must use -lpng to compile.
 */


#include <stdio.h>
#include <stdlib.h>
#include <png.h>

int simple_png_write_rgba(
  char *filename, 
  int  *rgba, int width, int height, 
  char *comment, 
  int   moffset, int noffset, int uncompressed) 
{
  FILE *fp;
  int i;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text *png_text_ptr=NULL;
  png_byte **rows;
  // open output file
  fp = fopen(filename, "wb");
  if (!fp) {
    fprintf(stderr, "error opening file %s.\n", filename);
    return -1;
  }

  // allocate png_write struct
  png_ptr =
      png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr) {
    fprintf(stderr, "failed png_create_write_struct()\n");
    return -1;
  }

  // allocate png_info struct
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    fprintf(stderr, "failed png_create_info_struct()\n");
    png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
    return -1;
  }

  // allocate and initialize rows
  rows = (png_byte**)malloc(height*sizeof(png_byte*));
  for (i = 0; i < height; i++) {
    rows[i] = (png_byte *) & (rgba[i * width]);
  }

  // alloc text for comment
  if (comment!=NULL) {
    png_text_ptr = (png_text*)malloc(sizeof(png_text));
    png_text_ptr->key="Comment";
    png_text_ptr->text=comment;
    png_text_ptr->compression=PNG_TEXT_COMPRESSION_zTXt;
    
    // set png comment fields
    png_set_text(png_ptr, info_ptr, png_text_ptr, 1);  
  }

  // set image offset values
  png_set_oFFs(png_ptr, info_ptr, 
      moffset, noffset, 
      PNG_OFFSET_PIXEL);  

  // prototype for file parameters
  //png_set_IHDR(png_ptr, info_ptr, width, height,
  //bit_depth, color_type, interlace_type,
  //compression_type, filter_type);
  png_set_IHDR(png_ptr, info_ptr, width, height,
               8, PNG_COLOR_TYPE_RGB_ALPHA,
               PNG_INTERLACE_NONE,
               PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

  // alpha=255 is always opaque
  // alpha=0   is totally transparent
  png_set_swap_alpha(png_ptr); // RGBA <-> ARGB
  // NOT NOW png_set_bgr(png_ptr);         // this seems to treat rgba correctly

  // write image header, data, then end 
  png_init_io(png_ptr, fp);
  if (uncompressed) { 
    png_set_compression_level(png_ptr, Z_NO_COMPRESSION); 
  }
  png_write_info(png_ptr, info_ptr);
  png_write_image(png_ptr, rows);
  png_write_end(png_ptr, info_ptr);

  // we must free any data allocated for info_ptr, 
  // such as comments, palette, or histogram, before 
  // the call to png_destroy_write_struct
  png_destroy_write_struct(&png_ptr, &info_ptr);

  // free allocated memory
  if (comment!=NULL) free(png_text_ptr);
  free(rows);

  fclose(fp);

  return 0;
}


/*  
 *  This is a SIMPLE interface to the png library
 *  for writing GA images.  the two bytes are
 *  packed into an assumed 16-bit short int in the 
 *  GA (greyscale alpha) order.
 *
 *  So we need the 
 *    filename, raw data ptr, width, height,
 *  and optional 
 *    comment (may be NULL)
 *    moffset, noffset (may be 0)
 *
 *  you must use -lpng to compile.
 */


int simple_png_write_ga(
  char *filename, 
  short int  *ga, int width, int height, 
  char *comment, 
  int   moffset, int noffset) 
{
  FILE *fp;
  int i;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text *png_text_ptr=NULL;
  png_byte **rows;
  // open output file
  fp = fopen(filename, "wb");
  if (!fp) {
    fprintf(stderr, "error opening file %s.\n", filename);
    return -1;
  }

  // allocate png_write struct
  png_ptr =
      png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr) {
    fprintf(stderr, "failed png_create_write_struct()\n");
    return -1;
  }

  // allocate png_info struct
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    fprintf(stderr, "failed png_create_info_struct()\n");
    png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
    return -1;
  }

  // allocate and initialize rows
  rows = (png_byte**)malloc(height*sizeof(png_byte*));
  for (i = 0; i < height; i++) {
    rows[i] = (png_byte *) & (ga[i*width]);
  }

  // alloc text for comment
  if (comment!=NULL) {
    png_text_ptr = (png_text*)malloc(sizeof(png_text));
    png_text_ptr->key="Comment";
    png_text_ptr->text=comment;
    png_text_ptr->compression=PNG_TEXT_COMPRESSION_zTXt;
    
    // set png comment fields
    png_set_text(png_ptr, info_ptr, png_text_ptr, 1);  
  }

  // set image offset values
  png_set_oFFs(png_ptr, info_ptr, 
      moffset, noffset, 
      PNG_OFFSET_PIXEL);  

  // prototype for file parameters
  //png_set_IHDR(png_ptr, info_ptr, width, height,
  //bit_depth, color_type, interlace_type,
  //compression_type, filter_type);
  png_set_IHDR(png_ptr, info_ptr, width, height,
               8, PNG_COLOR_TYPE_GRAY_ALPHA,
               PNG_INTERLACE_NONE,
               PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

  // alpha=255 is always opaque
  // alpha=0   is totally transparent
  png_set_swap_alpha(png_ptr); // GA <-> AG

  // write image header, data, then end 
  png_init_io(png_ptr, fp);
  png_write_info(png_ptr, info_ptr);
  png_write_image(png_ptr, rows);
  png_write_end(png_ptr, info_ptr);

  // we must free any data allocated for info_ptr, 
  // such as comments, palette, or histogram, before 
  // the call to png_destroy_write_struct
  png_destroy_write_struct(&png_ptr, &info_ptr);

  // free allocated memory
  if (comment!=NULL) free(png_text_ptr);
  free(rows);

  fclose(fp);

  return 0;
}

